import asyncio
import functools
import os
import signal
import sys
import time

import tornado.autoreload
import tornado.ioloop
import tornado.web
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string

from hurricane.amqp import logger
from hurricane.amqp.basehandler import _AMQPConsumer
from hurricane.amqp.worker import AMQPClient
from hurricane.metrics import StartupTimeMetric
from hurricane.server import make_probe_server


class Command(BaseCommand):

    """
    Starting a Tornado-powered Django AMQP 0-9-1 consumer.
    Implements consume command as a management command for django application.
    The new command can be called using ``python manage.py consume <arguments>``.
    Arguments:
        - ``--queue`` - the AMQP 0-9-1 queue to consume from
        - ``--exchange`` - the AMQP 0-9-1 exchange to declare
        - ``--amqp-port`` - the message broker connection port
        - ``--amqp-host`` - the host address of the message broker
        - ``--amqp-vhost`` - the virtual host of the message broker to use with this consumer
        - ``--handler`` - the Hurricane AMQP handler class (dotted path)
        - ``--startup-probe`` - the exposed path (default is /startup) for probes to check startup
        - ``--readiness-probe`` - the exposed path (default is /ready) for probes to check readiness
        - ``--liveness-probe`` - the exposed path (default is /alive) for probes to check liveness
        - ``--probe-port`` - the port for Tornado probe route to listen on
        - ``--req-queue-len`` - threshold of length of queue of request, which is considered for readiness probe
        - ``--no-probe`` - disable probe endpoint
        - ``--no-metrics`` - disable metrics collection
        - ``--autoreload`` - reload code on change
        - ``--debug`` - set Tornado's Debug flag
        - ``--reconnect`` - try to reconnect this client automatically as the broker is available again
    """

    help = "Start a Tornado-powered Django AMQP 0-9-1 consumer"

    def add_arguments(self, parser):

        """
        Defines arguments, that can be accepted with ``consume`` command.
        """

        parser.add_argument("--queue", type=str, default="", help="The AMQP 0-9-1 queue to consume from")
        parser.add_argument("--exchange", type=str, default="", help="The AMQP 0-9-1 exchange to declare")
        parser.add_argument(
            "--amqp-port",
            type=int,
            help="The message broker connection port",
        )
        parser.add_argument(
            "--amqp-host",
            type=str,
            help="The host address of the message broker",
        )
        parser.add_argument(
            "--amqp-vhost",
            type=str,
            help="The virtual host of the message broker to use with this consumer",
        )
        parser.add_argument("handler", type=str, help="The Hurricane AMQP handler class (dotted path)")
        parser.add_argument(
            "--liveness-probe",
            type=str,
            default="/alive",
            help="The exposed path (default is /alive) for probes to check liveness",
        )
        parser.add_argument(
            "--readiness-probe",
            type=str,
            default="/ready",
            help="The exposed path (default is /ready) for probes to check readiness",
        )
        parser.add_argument(
            "--startup-probe",
            type=str,
            default="/startup",
            help="The exposed path (default is /startup) for probes to check startup",
        )
        parser.add_argument(
            "--probe-port",
            type=int,
            default=8001,
            help="The port for Tornado probe route to listen on",
        )
        parser.add_argument("--req-queue-len", type=int, default=10, help="Length of the request queue")
        parser.add_argument("--no-probe", action="store_true", help="Disable probe endpoint")
        parser.add_argument("--no-metrics", action="store_true", help="Disable metrics collection")
        parser.add_argument("--autoreload", action="store_true", help="Reload code on change")
        parser.add_argument("--debug", action="store_true", help="Set Tornado's Debug flag")
        parser.add_argument(
            "--reconnect",
            action="store_true",
            help="Try to reconnect this client automatically as the broker is available again",
        )
        parser.add_argument(
            "--webhook-url",
            type=str,
            help="Url for webhooks",
        )

    def handle(self, *args, **options):

        """
        Defines functionalities for different arguments. After all arguments were processed, it starts the async event
        loop.
        """
        start_time = time.time()
        logger.info("Starting a Tornado-powered Django AMQP consumer")

        if options["autoreload"]:
            tornado.autoreload.start()
            logger.info("Autoreload was performed")

        # sanitize probe paths
        options["liveness_probe"] = f"/{options['liveness_probe'].lstrip('/')}"
        options["readiness_probe"] = f"/{options['readiness_probe'].lstrip('/')}"
        options["startup_probe"] = f"/{options['startup_probe'].lstrip('/')}"

        # set the probe routes
        if not options["no_probe"]:
            logger.info(f"Probe application running on port {options['probe_port']}")
            probe_application = make_probe_server(options, self.check)
            probe_application.listen(options["probe_port"])
        else:
            logger.info("No probe application running")

        # load connection data
        connection = {}
        if "amqp_host" in options and options["amqp_host"]:
            connection["amqp_host"] = options["amqp_host"]
        elif hasattr(settings, "AMQP_HOST"):
            connection["amqp_host"] = settings.AMQP_HOST
        elif os.getenv("AMQP_HOST"):
            connection["amqp_host"] = os.getenv("AMQP_HOST")
        else:
            raise CommandError(
                "The amqp host must not be empty: set it either as environment AMQP_HOST, "
                "in the django settings as AMQP_HOST or as optional argument --amqp-host"
            )
        if "amqp_port" in options and options["amqp_port"]:
            connection["amqp_port"] = options["amqp_port"]
        elif hasattr(settings, "AMQP_PORT"):
            connection["amqp_port"] = settings.AMQP_PORT
        elif os.getenv("AMQP_PORT"):
            connection["amqp_port"] = int(os.getenv("AMQP_PORT"))
        else:
            raise CommandError(
                "The amqp port must not be empty: set it either as environment AMQP_PORT, "
                "in the django settings as AMQP_PORT or as optional argument --amqp-port"
            )
        connection["amqp_vhost"] = "/"
        if "amqp_vhost" in options and options["amqp_vhost"]:
            connection["amqp_vhost"] = options["amqp_vhost"]
        elif hasattr(settings, "AMPQ_VHOST"):
            connection["amqp_vhost"] = settings.AMPQ_VHOST
        elif os.getenv("AMPQ_VHOST"):
            connection["amqp_vhost"] = os.getenv("AMPQ_VHOST")

        # load the handler class
        _amqp_consumer = import_string(options["handler"])
        if not issubclass(_amqp_consumer, _AMQPConsumer):
            logger.error(f"The type {_amqp_consumer} is not subclass of _AMQPConsumer")
            raise CommandError("Cannot start the consumer due to an implementation error")

        worker = AMQPClient(
            _amqp_consumer,
            queue_name=options["queue"],
            exchange_name=options["exchange"],
            amqp_host=connection["amqp_host"],
            amqp_port=connection["amqp_port"],
            amqp_vhost=connection["amqp_vhost"],
        )

        # prepate the io loop
        loop = asyncio.get_event_loop()

        def ask_exit(signame):
            logger.info(f"Received signal {signame}. Shutting down now.")
            loop.stop()
            sys.exit(0)

        for signame in ("SIGINT", "SIGTERM"):
            loop.add_signal_handler(getattr(signal, signame), functools.partial(ask_exit, signame))
        end_time = time.time()
        time_elapsed = end_time - start_time
        StartupTimeMetric.set(time_elapsed)
        worker.run(options["reconnect"])



from hurricane.metrics.registry import MetricsRegistry
from hurricane.metrics.requests import (
    HealthMetric,
    ReadinessMetric,
    RequestCounterMetric,
    RequestQueueLengthMetric,
    ResponseTimeAverageMetric,
    StartupTimeMetric,
)

registry = MetricsRegistry()

registry.register(RequestCounterMetric)
registry.register(RequestQueueLengthMetric)
registry.register(ResponseTimeAverageMetric)
registry.register(StartupTimeMetric)
registry.register(HealthMetric)
registry.register(ReadinessMetric)



import logging

access_log = logging.getLogger("hurricane.server.access")
app_log = logging.getLogger("hurricane.server.application")
logger = logging.getLogger("hurricane.server.general")
metrics_log = logging.getLogger("hurricane.server.metrics")



import asyncio
import functools
import socket
import time
from concurrent.futures.thread import ThreadPoolExecutor
from enum import Enum

import requests
from django.conf import settings
from requests import RequestException

from hurricane.server.loggers import logger


class WebhookStatus(Enum):
    FAILED = "failed"
    SUCCEEDED = "succeeded"


class Webhook:
    """
    Base class for webhooks in the registry. Run function initiates sending of webhook to the specified url.
    """

    code = None
    data = {}

    def __init__(self, code=None):
        if code:
            self.code = code

    @classmethod
    def get_from_registry(cls):

        """
        Getting webhook from registry using the code.
        """

        from hurricane.metrics import registry

        return registry.get(cls.code)

    def run(self, url: str, status: WebhookStatus, error_trace: str = None, close_loop: bool = False):

        """
        Initiates the sending of webhook in an asynchronous manner. Also specifies the callback of the async process,
        which handles the feedback and either logs success or failure of a webhook sending process.

        Parameters
        ----------
        url : Url, which webhook should be sent to
        status : can be either WebhookStatus.FAILED or WebhookStatus.SUCCEEDED depending on the success or failure of
        the process, which should be indicated by the webhook
        error_trace : specifies the error trace of the preceding failure
        close_loop : specifies, whether the main loop should be closed or be left running
        """

        self.set_traceback(error_trace)
        self.set_status(status)
        self.set_timestamp()
        self.set_podname()
        self.set_version()
        current_loop = asyncio.get_event_loop()
        executor = ThreadPoolExecutor(max_workers=1)
        fut = current_loop.run_in_executor(executor, self._send_webhook, self.get_message(), url, close_loop)
        # callback runs after run_in_executor is done
        callback_wrapper = functools.partial(self._callback_webhook_exception_check, url=url, close_loop=close_loop)
        fut.add_done_callback(callback_wrapper)

    def _send_webhook(self, data: dict, webhook_url: str, close_loop: bool):
        # sending webhook request to the specified url
        logger.info(f"Start sending {self.code} webhook to {webhook_url}")

        response = requests.post(webhook_url, timeout=5, json=data)
        response.raise_for_status()

    def set_traceback(self, error_trace: str):
        self.data["traceback"] = error_trace

    def set_timestamp(self):
        self.data["timestamp"] = int(time.time())

    def set_podname(self):
        self.data["podname"] = socket.gethostname()

    def set_status(self, status: WebhookStatus):
        self.data["status"] = status.value

    def set_version(self):
        if hasattr(settings, "HURRICANE_VERSION"):
            self.data["version"] = settings.HURRICANE_VERSION
        else:
            self.data["version"] = None

    def get_message(self):
        return self.data

    @staticmethod
    def _callback_webhook_exception_check(future: asyncio.Future, url: str, close_loop: bool):
        # checks if sending webhook had any failures, it indicates, that command was successfully executed
        # but sending webhook has failed
        try:
            future.result()
        except RequestException as e:
            logger.warning(f"Sending webhook to {url} has failed due to {e}")

        if close_loop:
            logger.info("Loop will be closed")
            current_loop = asyncio.get_event_loop()
            current_loop.stop()



import traceback

import tornado.web
from django.conf import settings
from django.core.management.base import SystemCheckError
from django.core.wsgi import get_wsgi_application
from django.db import OperationalError, connection

from hurricane.metrics import (
    HealthMetric,
    ReadinessMetric,
    RequestQueueLengthMetric,
    ResponseTimeAverageMetric,
    StartupTimeMetric,
)
from hurricane.server.loggers import logger
from hurricane.server.wsgi import HurricaneWSGIContainer
from hurricane.webhooks import LivenessWebhook, ReadinessWebhook
from hurricane.webhooks.base import WebhookStatus


class DjangoHandler(tornado.web.RequestHandler):

    """
    This handler transmits all standard requests to django application. Currently it uses WSGI Container based on
    tornado WSGI Container.
    """

    def initialize(self):
        """
        Initialization of Hurricane WSGI Container.
        """
        self.django = HurricaneWSGIContainer(self, get_wsgi_application())

    def prepare(self) -> None:
        """
        Transmitting incoming request to django application via WSGI Container.
        """
        self.django(self.request)
        self._finished = True
        self._log()
        self.on_finish()


class DjangoProbeHandler(tornado.web.RequestHandler):

    """
    Parent class for all specific probe handlers.
    """

    def compute_etag(self):
        return None

    def set_extra_headers(self, path):
        """
        Setting of extra headers for cache-control, namely: no-store, no-cache, must-revalidate and max-age=0. It
        means that information on requests and responses will not be stored.
        """
        self.set_header("Cache-Control", "no-store, no-cache, must-revalidate, max-age=0")

    def _check(self):
        """
        Checking application on several errors. Catches SystemCheckErrors of django system check framework and
        checks the connection to the database.
        """
        pass

    def get(self):
        """
        Get method, which runs the check.
        """
        self._check()

    def post(self):
        """
        Post method, which runs the check.
        """
        self._check()


class DjangoLivenessHandler(DjangoProbeHandler):
    """
    This handler runs with every call to the probe endpoint which is supposed to be used
    """

    def initialize(self, check_handler, webhook_url):
        self.check = check_handler
        self.liveness_webhook = webhook_url

    def _check(self):
        if StartupTimeMetric.get():
            got_exception = None
            try:
                self.check()
                if settings.DATABASES:
                    # once a connection has been established, this will be successful
                    # (even if the connection is gone later on)
                    connection.ensure_connection()
            except SystemCheckError as e:
                got_exception = traceback.format_exc()
                if settings.DEBUG:
                    self.write("django check error: " + str(e))
                else:
                    self.write("check error")
            except OperationalError as e:
                got_exception = traceback.format_exc()
                if settings.DEBUG:
                    self.write("django database error: " + str(e))
                else:
                    self.write("db error")

            else:
                if response_average_time := ResponseTimeAverageMetric.get():
                    self.write(
                        f"Average response time: {response_average_time:.2f}ms Request "
                        f"queue size: {RequestQueueLengthMetric.get()} Rx"
                    )
                else:
                    self.write("alive")
                if HealthMetric.get() is not True:
                    HealthMetric.set(True)
                    if self.liveness_webhook:
                        logger.info("Health metric changed to True. Liveness webhook with status succeeded triggered")
                        LivenessWebhook().run(url=self.liveness_webhook, status=WebhookStatus.SUCCEEDED)
            finally:
                if got_exception:
                    self.set_status(500)
                    if HealthMetric.get() is not False:
                        HealthMetric.set(False)
                        if self.liveness_webhook:
                            logger.info("Health metric changed to False. Liveness webhook with status failed triggered")
                            LivenessWebhook().run(
                                url=self.liveness_webhook, status=WebhookStatus.FAILED, error_trace=got_exception
                            )

        else:
            self.set_status(400)


class DjangoReadinessHandler(DjangoProbeHandler):
    """
    This handler runs with every call to the probe endpoint which is supposed to be used
    with Kubernetes 'Readiness Probes'. The DjangoCheckHandler calls Django's Check Framework which
    can be used to determine the application's health state during its operation.
    """

    def initialize(self, req_queue_len, webhook_url):
        self.request_queue_length = req_queue_len
        self.readiness_webhook = webhook_url

    def _check(self):
        if StartupTimeMetric.get() and RequestQueueLengthMetric.get() > self.request_queue_length:
            self.set_status(400)
            if ReadinessMetric.get() is not False:
                ReadinessMetric.set(False)
                if self.readiness_webhook:
                    logger.info("Readiness metric changed to False. Webhook with status failed triggered")
                    ReadinessWebhook().run(url=self.readiness_webhook, status=WebhookStatus.FAILED)
        elif StartupTimeMetric.get() and RequestQueueLengthMetric.get() <= self.request_queue_length:
            self.set_status(200)
            if ReadinessMetric.get() is not True:
                ReadinessMetric.set(True)
                if self.readiness_webhook:
                    logger.info("Readiness metric changed to True. Webhook with status succeeded triggered")
                    ReadinessWebhook().run(url=self.readiness_webhook, status=WebhookStatus.SUCCEEDED)
        else:
            self.set_status(400)


class DjangoStartupHandler(DjangoProbeHandler):
    """
    This handler runs with every call to the probe endpoint which is supposed to be used with Kubernetes
    'Startup Probes'. It returns 400 response for post and get requests, if StartupTimeMetric is not set, what means
    that the application is still in the startup phase. As soon as StartupTimeMetric is set, this handler returns 200
    response upon request, which indicates, that startup phase is finished and Kubernetes can now poll
    liveness/readiness probes.
    """

    def _check(self):
        if StartupTimeMetric.get():
            self.write(f"Startup was finished {StartupTimeMetric.get()}")
            self.set_status(200)
        else:
            self.set_status(400)



import tornado.web
from django.test import SimpleTestCase

from hurricane.testing.actors import HTTPClient, WebhookTestHandler
from hurricane.testing.drivers import HurricaneAMQPDriver, HurricaneServerDriver, HurricaneWebhookServerDriver


class HurricanBaseTest(SimpleTestCase):
    driver = None

    @property
    def probe_client(self):
        host, port = self.driver.get_server_host_port(probe_port=True)
        return HTTPClient(host, port)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.driver = cls.driver()


class HurricanServerTest(HurricanBaseTest):
    driver = HurricaneServerDriver

    @property
    def app_client(self):
        host, port = self.driver.get_server_host_port()
        return HTTPClient(host, port)

    @classmethod
    def tearDownClass(cls):
        cls.driver.stop_server()
        super().tearDownClass()

    @staticmethod
    def cycle_server(*args, **kwargs):
        def _cycle_server(function):
            def wrapper(self):
                if "args" in kwargs:
                    _args = kwargs["args"]
                else:
                    _args = None
                # run this hurricane server with coverage
                # default is True
                if "coverage" in kwargs:
                    coverage = kwargs["coverage"]
                else:
                    coverage = True
                self.driver.start_server(_args, coverage)
                try:
                    function(self)
                except Exception as e:
                    self.driver.stop_server()
                    raise e
                else:
                    self.driver.stop_server()

            return wrapper

        if len(args) == 1 and callable(args[0]):
            return _cycle_server(args[0])
        else:
            return _cycle_server


class HurricaneWebhookServerTest(HurricanServerTest):
    driver = HurricaneWebhookServerDriver

    @property
    def probe(self):
        host, port = self.driver.get_server_host_port(probe_port=True)
        return WebhookTestHandler()


class HurricaneAMQPTest(HurricanBaseTest):
    driver = HurricaneAMQPDriver

    @classmethod
    def tearDownClass(cls):
        cls.driver.stop_consumer()
        cls.driver.stop_amqp()
        super().tearDownClass()

    @staticmethod
    def cycle_consumer(*args, **kwargs):
        def _cycle_consumer(function):
            def wrapper(self):
                if "args" in kwargs:
                    _args = kwargs["args"]
                else:
                    _args = None
                # run this hurricane consumer with coverage
                # default is True
                if "coverage" in kwargs:
                    coverage = kwargs["coverage"]
                else:
                    coverage = True
                self.driver.start_amqp()
                host, port = self.driver.get_amqp_host_port()
                _args += ["--amqp-port", str(port), "--amqp-host", host]
                self.driver.start_consumer(_args, coverage)
                try:
                    function(self)
                except Exception as e:
                    self.driver.stop_consumer()
                    self.driver.stop_amqp()
                    raise e
                else:
                    self.driver.stop_consumer()
                    self.driver.stop_amqp()

            return wrapper

        if len(args) == 1 and callable(args[0]):
            return _cycle_consumer(args[0])
        else:
            return _cycle_consumer



from hurricane.webhooks.exceptions import WebhookCodeAlreadyRegistered


class WebhookRegistry:

    """
    Registering webhooks and storing them in a webhooks dictionary.
    """

    def __init__(self):
        self.webhooks = {}

    def register(self, webhook_cls):
        if webhook_cls.code in self.webhooks:
            raise WebhookCodeAlreadyRegistered(f"Webhook Code ({webhook_cls.code}) is already registered.")
        self.webhooks[webhook_cls.code] = webhook_cls()

    def unregister(self, webhook_cls):
        try:
            del self.webhooks[webhook_cls.code]
        except KeyError:
            # TODO warn about trying to unregister not registered metric
            pass

    def get(self, code):
        return self.webhooks[code]



class StoredMetric:
    """
    Base class for storing metrics in registry.
    """

    code = None
    value = None

    def __init__(self, code=None, initial=None):
        if code:
            self.code = code
        if initial:
            self.value = initial

    @classmethod
    def get_from_registry(cls):

        """
        Getting metric from registry using metric code.
        """

        from hurricane.metrics import registry

        return registry.get(cls.code)

    @classmethod
    def set(cls, value):

        """
        Setting new value for metric.
        """

        metric = cls.get_from_registry()
        metric.value = value

    @classmethod
    def get(cls):

        """
        Getting value of metric from registry.
        """

        metric = cls.get_from_registry()
        return metric.value


class CalculatedMetric:
    code = None

    def __init__(self, code=None):
        if code:
            self.code = code

    @classmethod
    def get_from_registry(cls):

        """
        Getting metric from registry using metric code.
        """

        from hurricane.metrics import registry

        return registry.get(cls.code)

    @classmethod
    def get(cls):

        """
        Getting value of metric from registry.
        """

        return cls.get_value(cls)

    @classmethod
    def get_value(cls):
        raise NotImplementedError


class CounterMetric(StoredMetric):

    """
    Metric, that can be incremented and decremented.
    """

    value = 0

    @classmethod
    def increment(cls):

        """
        Increment value to the metric.
        """

        cls.set(cls.get() + 1)

    @classmethod
    def decrement(cls):

        """
        Decrement value from the metric.
        """

        cls.set(cls.get() - 1)


class AverageMetric(StoredMetric):

    """
    Calculating average of a metric.
    """

    counter = 0
    value = 0

    @classmethod
    def add_value(cls, value):

        """
        Implements the running (online) average of a metric.
        """

        metric = cls.get_from_registry()
        metric.counter += 1
        metric.value = metric.value + (value - metric.value) / metric.counter



import logging

import tornado.ioloop

from hurricane.testing.actors import WebhookReceiverServer

if __name__ == "__main__":
    logging.info("Started webhook receiver server")
    app = WebhookReceiverServer().make_http_receiver_app()
    app.listen(8074)
    tornado.ioloop.IOLoop.current().start()



import asyncio
import functools
from typing import List

import pika
from pika.adapters.tornado_connection import TornadoConnection

from hurricane.amqp import logger


class _AMQPConsumer:

    """
    This is Hurricane's base AMQP consumer that will handle unexpected interactions
    with the message broker such as channel and connection closures.
    If the message broker closes the connection, this class will stop and indicate
    that reconnection is necessary.
    """

    EXCHANGE_TYPE = None

    def __init__(
        self,
        queue_name: str,
        exchange_name: str,
        host: str,
        port: int,
        vhost: str = None,
        username: str = None,
        password: str = None,
    ):
        """
        Create a new instance of the consumer class, passing in the AMQP
        URL used to connect to broker.
        """
        self.should_reconnect = False
        self.was_consuming = False

        self._queue_name = queue_name
        self._exchange_name = exchange_name
        self._host = host
        self._port = port
        self._vhost = vhost or "/"
        self._username = username
        self._password = password

        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._consuming = False
        self._prefetch_count = 1

    def connect(self) -> TornadoConnection:

        """
        This method connects to the broker, returning the connection handle.
        """

        logger.info(f"Connecting to {self._host}:{self._port}{self._vhost}")
        # set amqp credentials
        if self._username:
            credentials = pika.PlainCredentials(self._username, self._password)
            # set amqp connection parameters
            parameters = pika.ConnectionParameters(
                host=self._host,
                port=self._port,
                virtual_host=self._vhost,
                credentials=credentials,
            )
        else:
            parameters = pika.ConnectionParameters(
                host=self._host,
                port=self._port,
                virtual_host=self._vhost,
            )

        # connect
        connection = TornadoConnection(
            parameters=parameters,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
        )
        return connection

    def close_connection(self):
        self._consuming = False
        if self._connection.is_closing or self._connection.is_closed:
            logger.info("Connection is closing or already closed")
        else:
            logger.info("Closing connection")
            self._connection.close()

    def on_connection_open(self, _unused_connection: pika.SelectConnection):

        """
        This method is called by pika once the connection to the broker has
        been established.
        """

        logger.info("Connection opened")
        self.open_channel()

    def on_connection_open_error(self, _unused_connection: pika.SelectConnection, err: Exception):

        """
        This method is called by pika if the connection to the broker
        can't be established.
        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error
        """

        logger.error(f"Connection open failed: {err}")
        self.reconnect()

    def on_connection_closed(self, _unused_connection: pika.SelectConnection, reason: Exception):

        """
        This method is invoked by pika when the connection to the broker is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        the broker if it disconnects.
        """

        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            logger.warning("Connection closed, reconnect necessary: %s", reason)
            self.reconnect()

    def reconnect(self):

        """
        Will be invoked if the connection can't be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.
        """

        self.should_reconnect = True
        self.stop()

    def open_channel(self):

        """
        Open a new channel with the broker by issuing the Channel.Open RPC
        command. When the broker responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """

        logger.info("Creating a new channel")
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel: pika.channel.Channel):

        """
        This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.
        """

        logger.info("Channel opened")
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self._exchange_name)

    def add_on_channel_close_callback(self):

        """
        This method tells pika to call the on_channel_closed method if
        the broker unexpectedly closes the channel.
        """

        logger.info("Adding channel close callback")
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel: pika.channel.Channel, reason: Exception):

        """
        Invoked by pika when the broker unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        """

        logger.warning("Channel %i was closed: %s", channel, reason)
        self.close_connection()

    def setup_exchange(self, exchange_name: str) -> None:

        """
        Setup the exchange on the broker by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.
        """

        logger.info(f"Declaring exchange: {exchange_name}")
        # Note: using functools.partial is not required, it is demonstrating
        # how arbitrary data can be passed to the callback when it is called
        cb = functools.partial(self.on_exchange_declareok, userdata=exchange_name)
        self._channel.exchange_declare(exchange=exchange_name, exchange_type=self.EXCHANGE_TYPE, callback=cb)

    def on_exchange_declareok(self, _unused_frame: pika.frame.Method, userdata: str) -> None:

        """
        Invoked by pika when the broker has finished the Exchange.Declare RPC
        command.
        """

        logger.info("Exchange declared: %s", userdata)
        self.setup_queue(self._queue_name)

    def setup_queue(self, queue_name: str) -> None:

        """
        Setup the queue on the broker by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.
        :param str|unicode queue_name: The name of the queue to declare.
        """

        logger.info(f"Declaring queue {queue_name}")
        cb = functools.partial(self.on_queue_declareok, userdata=queue_name)
        self._channel.queue_declare(queue=queue_name, callback=cb)

    def get_routing_keys(self, queue_name: str) -> List[str]:

        """
        Generate a list of binding keys for this queue. This method will
        be called from on_queue_declareok in order to bind the declared queue
        on to one or multiple routing keys.
        """

        return []

    def on_queue_declareok(self, _unused_frame: pika.frame.Method, userdata: str) -> None:

        """
        Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.
        """

        queue_name = userdata
        logger.info(f"Binding to {queue_name}")
        routing_keys = self.get_routing_keys(queue_name)
        if len(routing_keys) == 0:
            # no routing key applicable
            cb = functools.partial(self.on_bindok, queue_name=queue_name)
            self._channel.queue_bind(queue_name, exchange=self._exchange_name, callback=cb)
        else:
            for routing_key in routing_keys:
                cb = functools.partial(self.on_bindok, queue_name=queue_name, routing_key=routing_key)
                self._channel.queue_bind(queue_name, routing_key=routing_key, exchange=self._exchange_name, callback=cb)

    def on_bindok(self, _unused_frame: pika.frame.Method, queue_name: str, routing_key: str = None):

        """
        Invoked by pika when the Queue.Bind method has completed. At this
        point we will set the prefetch count for the channel.
        """

        if routing_key:
            logger.info(f"Queue bound: {queue_name} with routing key {routing_key}")
        else:
            logger.info(f"Queue bound: {queue_name}")
        self.set_qos()

    def set_qos(self) -> None:

        """
        This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before the broker will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.
        """

        self._channel.basic_qos(prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok)

    def on_basic_qos_ok(self, _unused_frame) -> None:

        """
        Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.
        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame
        """

        logger.info(f"QOS set to: {self._prefetch_count}")
        self.start_consuming()

    def start_consuming(self) -> None:

        """
        This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if the broker
        cancels the consumer.
        """

        logger.info("Issuing consumer")
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(self._queue_name, self.on_message)
        self.was_consuming = True
        self._consuming = True

    def add_on_cancel_callback(self) -> None:

        """
        Add a callback that will be invoked if the broker cancels the consumer
        for some reason. If the broker does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.
        """

        logger.info("Adding consumer cancellation callback")
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame: pika.frame.Method) -> None:

        """
        Invoked by pika when the broker sends a Basic.Cancel for a consumer
        receiving messages.
        """

        logger.info(f"Consumer was cancelled remotely, shutting down: {method_frame}")
        if self._channel:
            self._channel.close()

    def on_message(
        self,
        _unused_channel: pika.channel.Channel,
        basic_deliver: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: str,
    ) -> None:

        """
        Invoked by pika when a message is delivered from the broker.
        """

        logger.error(
            "Received message # %s from %s",
            basic_deliver.delivery_tag,
            properties.app_id,
        )
        self.reject_message(basic_deliver.delivery_tag)
        raise NotImplementedError("The on_message method must be implemented by a handler class")

    def acknowledge_message(self, delivery_tag) -> None:

        """
        Acknowledge the message delivery from the broker by sending a
        Basic.Ack RPC method for the delivery tag.
        :param int delivery_tag: The delivery tag from the Basic.Deliver frame
        """

        logger.info("Acknowledging message %s", delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def reject_message(self, delivery_tag, requeue: bool = False) -> None:

        """
        Reject the message delivery from the broker.
        """

        logger.info("Rejecting message %s", delivery_tag)
        self._channel.basic_nack(delivery_tag, requeue=requeue)

    def stop_consuming(self) -> None:

        """
        Tell the broker that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """

        if self._channel:
            logger.info("Sending a Basic.Cancel command to the broker")
            cb = functools.partial(self.on_cancelok, userdata=self._consumer_tag)
            self._channel.basic_cancel(self._consumer_tag, cb)

    def on_cancelok(self, _unused_frame: pika.frame.Method, userdata: str) -> None:

        """
        This method is invoked by pika when the broker acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.
        """

        self._consuming = False
        logger.info(f"The broker acknowledged the cancellation of the consumer: {userdata}")
        self.close_channel()

    def close_channel(self) -> None:

        """
        Call to close the channel cleanly by issuing the
        Channel.Close RPC command.
        """

        logger.info("Closing the channel")
        self._channel.close()

    def run(self) -> None:

        """
        Run the consumer by connecting to the broker and then
        starting the IOLoop to block and allow the SelectConnection to operate.
        """

        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self) -> None:

        """
        Cleanly shutdown the connection by stopping the consumer.
        """

        if not self._closing:
            self._closing = True
            logger.info("Stopping")
            if self._consuming:
                self.stop_consuming()
                logger.info("Start called")
                self._connection.ioloop.stop()
            else:
                self._connection.ioloop.stop()
            logger.info("Stopped consuming")


class TopicHandler(_AMQPConsumer):

    """
    This handler implements Hurricane's base AMQP consumer that handles unexpected interactions
    with the message broker such as channel and connection closures. The ``EXCHANGE_TYPE`` is *topic*.
    """

    EXCHANGE_TYPE = "topic"



import socket
import subprocess
from queue import Empty, Queue
from threading import Thread
from time import sleep
from typing import List, Tuple, Union

import docker

from hurricane.testing.actors import TestPublisher

MANAGE_FILE = "manage.py"


class BusyPortException(Exception):
    pass


class HurricaneBaseDriver(object):
    proc = None
    log_lines = []
    base_command = []
    coverage_base_command = []
    test_string = ""
    ports = [8000, 8001]

    def __init__(self):
        for port in self.ports:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                try:
                    sock.bind(("127.0.0.1", port))
                except OSError:
                    raise BusyPortException(f"Port {port} already in use.")

    def get_server_host_port(self, probe_port=False) -> Union[Tuple[str, int], Tuple[None, None]]:
        port = self.probe_port if probe_port else self.port
        if self.proc:
            return "localhost", port
        else:
            return None, None

    def get_output(self, read_all=False) -> Tuple[str, str]:
        if read_all:
            while True:
                try:
                    line = self.q.get(timeout=0.5)
                    self.log_lines.append(line)
                except Empty:
                    break
            return "".join(self.log_lines), ""
        else:
            try:
                line = self.q.get(timeout=1)
                if line:
                    return line, ""
            except Empty:
                pass

    def _start(self, params: List[str] = None, coverage: bool = True) -> None:
        self.log_lines = []
        base_command = self.coverage_base_command if coverage else self.base_command

        def enqueue_stdout(proc, queue):
            out = proc.stdout
            for line in iter(out.readline, b""):
                queue.put(line.decode("utf-8"))
            out.close()

        def enqueue_stderr(proc, queue):
            out = proc.stderr
            for line in iter(out.readline, b""):
                queue.put(line.decode("utf-8"))
            out.close()

        params = params if params else []

        base_command = base_command + params
        if "--port" in params:
            self.port = int(params[params.index("--port") + 1])
        else:
            self.port = 8000

        if "--probe-port" in params:
            self.probe_port = int(params[params.index("--probe-port") + 1])
        else:
            self.probe_port = 8001

        self.proc = subprocess.Popen(base_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.q = Queue()
        self.t_stderr = Thread(target=enqueue_stderr, args=(self.proc, self.q))
        self.t_stdout = Thread(target=enqueue_stdout, args=(self.proc, self.q))
        self.t_stderr.daemon = True
        self.t_stderr.start()
        self.t_stdout.daemon = True
        self.t_stdout.start()
        # wait a maximum of 1 second
        for _ in range(10):
            if self.proc:
                out, err = self.get_output(read_all=True)
            else:
                out = ""
            if self.test_string in out:
                break
            sleep(0.1)

    def _stop(self):
        if self.proc:
            self.proc.terminate()


class HurricaneServerDriver(HurricaneBaseDriver):
    coverage_base_command = [
        "coverage",
        "run",
        "--source=hurricane/",
        MANAGE_FILE,
        "serve",
    ]
    base_command = ["python", MANAGE_FILE, "serve"]
    test_string = "Tornado-powered Django web server"

    def start_server(self, params: dict = None, coverage: bool = True) -> None:
        self._start(params, coverage)

    def stop_server(self) -> None:
        self._stop()


class HurricaneWebhookServerDriver(HurricaneBaseDriver):
    ports = [8040, 8041]
    coverage_base_command = [
        "coverage",
        "run",
        "-m",
        "--source=hurricane",
        "hurricane.testing.start_webhook_receiver",
    ]
    base_command = ["python", "-m", "hurricane.testing.start_webhook_receiver"]
    test_string = "Started webhook receiver server"

    def start_server(self, params: dict = None, coverage: bool = True) -> None:
        self._start(params, coverage)

    def stop_server(self) -> None:
        self._stop()


class HurricaneAMQPDriver(HurricaneBaseDriver):
    coverage_base_command = [
        "coverage",
        "run",
        "--source=hurricane/",
        MANAGE_FILE,
        "consume",
    ]
    base_command = ["python", MANAGE_FILE, "consume"]
    test_string = "Starting a Tornado-powered Django AMQP consumer"
    ports = [5672, 8000, 8001]

    def start_amqp(self) -> None:
        client = docker.from_env()
        if hasattr(self, "container") and self.container:
            c = client.containers.run(
                "quay.io/blueshoe/rabbitmq3.8-alpine",
                auto_remove=True,
                detach=True,
                ports={"5672": ("127.0.0.1", self._temp_port)},
            )
        else:
            c = client.containers.run(
                "quay.io/blueshoe/rabbitmq3.8-alpine",
                auto_remove=True,
                detach=True,
                ports={"5672": ("127.0.0.1", None)},
            )
        self.container = client.containers.get(c.id)
        # busy wait for rabbitmq to come up (timeout 10 seconds)
        for _ in range(20):
            if "Ready to start client connection listeners" in self.container.logs().decode("utf-8"):
                break
            else:
                sleep(0.5)

    def get_test_publisher(self, vhost="/"):
        host, port = self.get_amqp_host_port()
        return TestPublisher(host, port, vhost)

    def start_consumer(self, params: List[str] = None, coverage: bool = True) -> None:
        self._start(params, coverage)

    def stop_amqp(self) -> None:
        if hasattr(self, "container") and self.container:
            try:
                self.container.kill()
                delattr(self, "container")
            except Exception:
                # this container is potentially already stopped
                delattr(self, "container")

    def halt_amqp(self) -> None:
        if hasattr(self, "container") and self.container:
            try:
                _, self._temp_port = self.get_amqp_host_port()
                self.container.kill()
            except Exception:
                # this container is potentially already stopped
                pass

    def stop_consumer(self) -> None:
        self._stop()

    def _get_port(self):
        if hasattr(self, "container") and self.container:
            self._temp_port = self.container.attrs["NetworkSettings"]["Ports"]["5672/tcp"][0]["HostPort"]
            return self._temp_port
        return None

    def get_amqp_host_port(self) -> Union[Tuple[str, int], Tuple[None, None]]:
        if port := self._get_port():
            return "127.0.0.1", port
        else:
            return None, None



import asyncio
import time
import traceback
from typing import Callable

import tornado
from django.conf import settings
from django.core.management import call_command

from hurricane.metrics import RequestCounterMetric, ResponseTimeAverageMetric, StartupTimeMetric
from hurricane.server.django import DjangoHandler, DjangoLivenessHandler, DjangoReadinessHandler, DjangoStartupHandler
from hurricane.server.loggers import access_log, logger
from hurricane.webhooks import StartupWebhook
from hurricane.webhooks.base import WebhookStatus


class HurricaneApplication(tornado.web.Application):
    def __init__(self, *args, **kwargs):
        self.collect_metrics = True
        if "metrics" in kwargs:
            self.collect_metrics = kwargs["metrics"]
        super(HurricaneApplication, self).__init__(*args, **kwargs)

    def log_request(self, handler: DjangoHandler) -> None:
        """Writes a completed HTTP request to the logs. """
        if handler.get_status() < 400:
            log_method = access_log.info
        elif handler.get_status() < 500:
            log_method = access_log.warning
        else:
            log_method = access_log.error
        request_time = 1000.0 * handler.request.request_time()
        log_method(
            "%d %s %.2fms",
            handler.get_status(),
            handler._request_summary(),
            request_time,
        )
        if self.collect_metrics:
            RequestCounterMetric.increment()
            ResponseTimeAverageMetric.add_value(request_time)


def make_probe_server(options, check_func):
    """ create probe route application """
    handlers = [
        (
            options["liveness_probe"],
            DjangoLivenessHandler,
            {"check_handler": check_func, "webhook_url": options["webhook_url"]},
        ),
        (
            options["readiness_probe"],
            DjangoReadinessHandler,
            {"req_queue_len": options["req_queue_len"], "webhook_url": options["webhook_url"]},
        ),
        (options["startup_probe"], DjangoStartupHandler),
    ]
    return HurricaneApplication(handlers, debug=options["debug"], metrics=False)


def make_http_server(options, check_func, include_probe=False):
    """ create all routes for this application """
    if include_probe:
        handlers = [
            (options["liveness_probe"], DjangoLivenessHandler, {"check_handler": check_func}),
            (options["readiness_probe"], DjangoReadinessHandler, {"req_queue_len": options["req_queue_len"]}),
            (options["startup_probe"], DjangoStartupHandler),
        ]
    else:
        handlers = []
    # if static file serving is enabled
    if options["static"]:
        logger.info(f"Serving static files under {settings.STATIC_URL} from {settings.STATIC_ROOT}")
        handlers.append(
            (
                f"{settings.STATIC_URL}(.*)",
                tornado.web.StaticFileHandler,
                {"path": settings.STATIC_ROOT},
            )
        )
    # if media file serving is enabled
    if options["media"]:
        logger.info(f"Serving media files under {settings.MEDIA_URL} from {settings.MEDIA_ROOT}")
        handlers.append(
            (
                f"{settings.MEDIA_URL}(.*)",
                tornado.web.StaticFileHandler,
                {"path": settings.MEDIA_ROOT},
            )
        )

    # append the django routing system
    handlers.append((".*", DjangoHandler))
    return HurricaneApplication(handlers, debug=options["debug"], metrics=not options["no_metrics"])


def make_http_server_and_listen(start_time: float, options: dict, check: Callable, include_probe: bool) -> None:
    logger.info(f"Starting HTTP Server on port {options['port']}")
    django_application = make_http_server(options, check, include_probe)
    django_application.listen(options["port"])
    end_time = time.time()
    time_elapsed = end_time - start_time
    # if startup time metric value is set - startup process is finished
    StartupTimeMetric.set(time_elapsed)
    logger.info(f"Startup time is {time_elapsed} seconds")


def command_task(main_loop: asyncio.unix_events.SelectorEventLoop, callback: Callable, commands: list) -> None:
    logger.info("Starting execution of management commands")

    for command in commands:

        # split a command string to get command options
        command_split = command[0].split()
        logger.info(f"Starting execution of command {command_split[0]} with arguments {command_split[1:]}")
        start_time_command = time.time()
        # call management command
        call_command(*command_split)
        end_time_command = time.time()

        logger.info(f"Command {command_split[0]} was executed in {end_time_command - start_time_command} seconds")
    # start http server in the main loop
    main_loop.call_soon_threadsafe(callback)


def callback_command_exception_check(future: asyncio.Future, webhook_url: str = None) -> None:
    # checks if there were any exceptions in the executor and if any stops the loop
    # if exceptions occured, it means, that some or one of the commands have failed
    if webhook_url:
        try:
            future.result()
        except Exception as e:
            logger.error("Execution of management commands has failed. Webhook should be sent")
            # printing full tracestack
            error_trace = traceback.format_exc()
            logger.error(e)
            logger.error(traceback.print_exc())
            logger.info("Webhook with a status failed has been initiated")
            StartupWebhook().run(url=webhook_url, error_trace=error_trace, close_loop=True, status=WebhookStatus.FAILED)
        else:
            logger.info("Execution of management commands was successful. Webhook should be sent")
            logger.info("Webhook with a status succeeded has been initiated")
            StartupWebhook().run(url=webhook_url, status=WebhookStatus.SUCCEEDED)
    else:
        try:
            future.result()
        except Exception as e:
            logger.error("Execution of management command has failed. Webhook is not set")
            logger.error(e)
            logger.error(traceback.print_exc())
            current_loop = asyncio.get_event_loop()
            current_loop.stop()
        else:
            logger.info("Execution of management commands was successful. Webhook is not set")



from hurricane.metrics.exceptions import MetricIdAlreadyRegistered


class MetricsRegistry:

    """
    Registering metrics and storing them in a metrics dictionary.
    """

    def __init__(self):
        self.metrics = {}

    def register(self, metric_cls):
        if metric_cls.code in self.metrics:
            raise MetricIdAlreadyRegistered(f"Metric ID ({metric_cls.code}) is already registered.")
        self.metrics[metric_cls.code] = metric_cls()

    def unregister(self, metric_cls):
        try:
            del self.metrics[metric_cls.code]
        except KeyError:
            # TODO warn about trying to unregister not registered metric
            pass

    def get(self, code):
        return self.metrics[code]



import asyncio

from hurricane.metrics.base import AverageMetric, CalculatedMetric, CounterMetric, StoredMetric


class RequestCounterMetric(CounterMetric):

    """
    Defines request counter metric with corresponding metric code.
    """

    code = "request_counter"


class ResponseTimeAverageMetric(AverageMetric):

    """
    Defines response time average metric with corresponding metric code.
    """

    code = "response_time_average"


class RequestQueueLengthMetric(CalculatedMetric):

    """
    Defines request queue length metric with corresponding metric code.
    """

    code = "request_queue_length"

    def get_value(self):

        """
        Getting length of the asyncio queue of all tasks.
        """

        return len(asyncio.all_tasks())


class StartupTimeMetric(StoredMetric):
    code = "startup_time"


class HealthMetric(StoredMetric):
    code = "health"


class ReadinessMetric(StoredMetric):
    code = "readiness"



from hurricane.webhooks.base import Webhook


class StartupWebhook(Webhook):
    code = "startup"

    def __init__(self):
        super(Webhook, self).__init__()
        self.data = {"info": "Startup webhook", "type": self.code}


class LivenessWebhook(Webhook):
    code = "liveness"

    def __init__(self):
        super(Webhook, self).__init__()
        self.data = {"info": "Liveness webhook", "type": self.code}


class ReadinessWebhook(Webhook):
    code = "readiness"

    def __init__(self):
        super(Webhook, self).__init__()
        self.data = {"info": "Readiness webhook", "type": self.code}



from hurricane.webhooks.registry import WebhookRegistry
from hurricane.webhooks.webhook_types import LivenessWebhook, ReadinessWebhook, StartupWebhook

webhook_registry = WebhookRegistry()
webhook_registry.register(StartupWebhook)
webhook_registry.register(LivenessWebhook)
webhook_registry.register(ReadinessWebhook)



import os
import time
from typing import Type

import tornado.ioloop
from django.conf import settings

from hurricane.amqp import logger
from hurricane.amqp.basehandler import _AMQPConsumer


class AMQPClient(object):

    """
    This is the AMQP Client that will reconnect, if the nested handler instance indicates that a reconnect is necessary.
    """

    def __init__(
        self,
        consumer_klass: Type[_AMQPConsumer],
        queue_name: str,
        exchange_name: str,
        amqp_host: str,
        amqp_port: int,
        amqp_vhost: str,
    ):
        self._reconnect_delay = 0

        # load user
        if hasattr(settings, "AMQP_USER"):
            user = settings.AMQP_USER
        elif os.getenv("AMQP_USER"):
            user = os.getenv("AMQP_USER")
        else:
            user = None
        # load password
        if hasattr(settings, "AMQP_PASSWORD"):
            password = settings.AMQP_PASSWORD
        elif os.getenv("AMQP_PASSWORD"):
            password = os.getenv("AMQP_PASSWORD")
        else:
            password = None

        self._consumer_args = (queue_name, exchange_name, amqp_host, amqp_port, amqp_vhost, user, password)
        self._consumer_klass = consumer_klass
        self._consumer = self._consumer_klass(*self._consumer_args)

    def run(self, reconnect: bool = False) -> None:

        """
        If reconnect is True, AMQP consumer is running in auto-connect mode.
        In this case consumer will be executed. If any exception occurs, consumer will be disconnected and after some
        delay will be reconnected. Then consumer will be restarted. KeyboardInterrupt exception is handled
        differently and stops consumer. In this case IOLoop will be terminated.

        If reconnect is false, consumer will be started, but no exceptions and interruptions will be tolerated.
        """

        if reconnect:
            logger.info("AMQP consumer running in auto-reconnect mode")
            while True:
                try:
                    self._consumer.run()
                except KeyboardInterrupt:
                    self._consumer.stop()
                    break
                self._maybe_reconnect()
        else:
            self._consumer.run()
        logger.info("Terminating Hurricane AMQP client")
        loop = tornado.ioloop.IOLoop.current()
        loop.stop()

    def _maybe_reconnect(self) -> None:
        if self._consumer.should_reconnect:
            self._consumer.stop()
            reconnect_delay = self._get_reconnect_delay()
            logger.warning("Reconnecting after %d seconds", reconnect_delay)
            time.sleep(reconnect_delay)
            self._consumer = self._consumer_klass(*self._consumer_args)

    def _get_reconnect_delay(self):
        if self._consumer.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay



import asyncio
import functools
import signal
import time
from concurrent.futures.thread import ThreadPoolExecutor

import tornado.autoreload
import tornado.web
import tornado.wsgi
from django.core.management.base import BaseCommand

from hurricane.server import (
    callback_command_exception_check,
    command_task,
    logger,
    make_http_server_and_listen,
    make_probe_server,
)
from hurricane.webhooks import StartupWebhook
from hurricane.webhooks.base import WebhookStatus


class Command(BaseCommand):

    """
    Start a Tornado-powered Django web server.
    Implements serve command as a management command for django application.
    The new command can be called using ``python manage.py server <arguments>``.
    It also can take command arguments, which are python django management commands and will be executed asynchronously.
    Upon successful execution of management commands, application server will be started. During execution of management
    commands probe server can be polled, in particular startup probe, which will respond with a status 400.
    Arguments:
        - ``--static`` - serve collected static files
        - ``--media`` - serve media files
        - ``--autoreload`` - reload code on change
        - ``--debug`` - set Tornado's Debug flag
        - ``--port`` - the port for Tornado to listen on
        - ``--startup-probe`` - the exposed path (default is /startup) for probes to check startup
        - ``--readiness-probe`` - the exposed path (default is /ready) for probes to check readiness
        - ``--liveness-probe`` - the exposed path (default is /alive) for probes to check liveness
        - ``--probe-port`` - the port for Tornado probe route to listen on
        - ``--req-queue-len`` - threshold of length of queue of request, which is considered for readiness probe
        - ``--no-probe`` - disable probe endpoint
        - ``--no-metrics`` - disable metrics collection
        - ``--command`` - repetitive command for adding execution of management commands before serving
        - ``--webhook-url``- If specified, webhooks will be sent to this url
    """

    help = "Start a Tornado-powered Django web server"

    def add_arguments(self, parser):
        """
        Defines arguments, that can be accepted with ``serve`` command.
        """
        parser.add_argument("--static", action="store_true", help="Serve collected static files")
        parser.add_argument("--media", action="store_true", help="Serve media files")
        parser.add_argument("--autoreload", action="store_true", help="Reload code on change")
        parser.add_argument("--debug", action="store_true", help="Set Tornado's Debug flag")
        parser.add_argument("--port", type=int, default=8000, help="The port for Tornado to listen on")
        parser.add_argument(
            "--liveness-probe",
            type=str,
            default="/alive",
            help="The exposed path (default is /alive) for probes to check liveness",
        )
        parser.add_argument(
            "--readiness-probe",
            type=str,
            default="/ready",
            help="The exposed path (default is /ready) for probes to check readiness",
        )
        parser.add_argument(
            "--startup-probe",
            type=str,
            default="/startup",
            help="The exposed path (default is /startup) for probes to check startup",
        )
        parser.add_argument(
            "--probe-port",
            type=int,
            help="The port for Tornado probe route to listen on",
        )
        parser.add_argument("--req-queue-len", type=int, default=10, help="Length of the request queue")
        parser.add_argument("--no-probe", action="store_true", help="Disable probe endpoint")
        parser.add_argument("--no-metrics", action="store_true", help="Disable metrics collection")
        parser.add_argument("--command", type=str, action="append", nargs="+")
        parser.add_argument(
            "--webhook-url",
            type=str,
            help="Url for webhooks",
        )

    def handle(self, *args, **options):
        """
        Defines functionalities for different arguments. After all arguments were processed, it starts the async event
        loop.
        """
        start_time = time.time()
        logger.info("Tornado-powered Django web server")

        if options["autoreload"]:
            tornado.autoreload.start()
            logger.info("Autoreload was performed")

        # set the probe port
        # the probe port by default is supposed to run the next port of the application
        probe_port = options["probe_port"] if options["probe_port"] else options["port"] + 1

        # sanitize probe paths
        options["liveness_probe"] = f"/{options['liveness_probe'].lstrip('/')}"
        options["readiness_probe"] = f"/{options['readiness_probe'].lstrip('/')}"
        options["startup_probe"] = f"/{options['startup_probe'].lstrip('/')}"

        # set the probe routes
        # if probe port is set to the application's port, add it to the application's routes
        include_probe = False
        if not options["no_probe"]:
            if probe_port != options["port"]:
                logger.info(
                    f"Starting probe application running on port {probe_port} with route liveness-probe: "
                    f"{options['liveness_probe']}, readiness-probe: {options['readiness_probe']}, "
                    f"startup-probe: {options['startup_probe']}"
                )
                probe_application = make_probe_server(options, self.check)
                probe_application.listen(probe_port)
            else:
                include_probe = True
                logger.info(
                    f"Starting probe application with routes liveness-probe: {options['liveness_probe']}, "
                    f"readiness-probe: {options['readiness_probe']}, startup-probe: {options['startup_probe']} "
                    f"running integrated on port {probe_port}"
                )

        else:
            logger.info("No probe application running")

        loop = asyncio.get_event_loop()

        if options["command"]:
            # command_task is a function that executes management commands
            # make_http_server_and_listen is a callback, further parameters are parameters of the callback function
            executor = ThreadPoolExecutor(max_workers=1)
            make_http_server_wrapper = functools.partial(
                make_http_server_and_listen,
                start_time=start_time,
                options=options,
                check=self.check,
                include_probe=include_probe,
            )
            future = loop.run_in_executor(executor, command_task, loop, make_http_server_wrapper, options["command"])
            if options["webhook_url"]:
                # wrap callback function to use additional variables
                callback_wrapper_command_exception_check = functools.partial(
                    callback_command_exception_check, webhook_url=options["webhook_url"]
                )
                # callback runs after run_in_executor is done
                future.add_done_callback(callback_wrapper_command_exception_check)
            else:
                future.add_done_callback(callback_command_exception_check)
        else:
            make_http_server_and_listen(start_time, options, self.check, include_probe)
            if options["webhook_url"]:
                StartupWebhook().run(url=options["webhook_url"], status=WebhookStatus.SUCCEEDED)

        def ask_exit(signame):
            logger.info(f"Received signal {signame}. Shutting down now.")
            loop.stop()

        for signame in ("SIGINT", "SIGTERM"):
            loop.add_signal_handler(getattr(signal, signame), functools.partial(ask_exit, signame))

        loop.run_forever()



# -*- coding: utf-8 -*-
import os

from setuptools import find_packages, setup

base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, "VERSION")) as f:
    VERSION = f.read()

with open(os.path.join(base_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

DESCRIPTION = "Hurricane is an initiative to fit Django perfectly with Kubernetes."


setup(
    name="django-hurricane",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=VERSION,
    install_requires=[
        "tornado~=6.1",
        "Django>=2.2",
        "pika~=1.1.0",
    ],
    python_requires="~=3.8",
    packages=[
        "hurricane",
        "hurricane.amqp",
        "hurricane.metrics",
        "hurricane.server",
        "hurricane.webhooks",
        "hurricane.testing",
        "hurricane.management",
        "hurricane.management.commands",
    ],
    author="Michael Schilonka",
    author_email="michael@blueshoe.de",
    include_package_data=True,
    url="https://github.com/Blueshoe/django-hurricane",
    project_urls={
        "Source": "https://github.com/Blueshoe/django-hurricane",
        "Documentation": "https://django-hurricane.readthedocs.io/en/latest/",
        "Bug Tracker": "https://github.com/Blueshoe/django-hurricane/issues",
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
)



import http
import json
import logging

import pika
import tornado.ioloop
import tornado.web

logging.basicConfig(level=logging.INFO)


class HTTPClient(object):
    class Response(dict):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__dict__ = self

    def __init__(self, host: str, port: int):
        self.conn = http.client.HTTPConnection(host=host, port=port)

    def get(self, path: str) -> Response:
        self.conn.request("GET", path)
        res = self.conn.getresponse()
        data = res.read()
        return self.Response({"status": res.status, "text": data.decode("utf-8")})

    def head(self, path: str) -> Response:
        self.conn.request("HEAD", path)
        res = self.conn.getresponse()
        data = res.read()
        return self.Response({"status": res.status, "text": data.decode("utf-8")})

    def post(self, path: str, data: dict) -> Response:
        self.conn.request("POST", path, json.dumps(data).encode())
        res = self.conn.getresponse()
        data = res.read()
        return self.Response({"status": res.status, "text": data.decode("utf-8")})


class TestPublisher(object):
    def __init__(self, host, port, vhost="/"):
        parameters = pika.ConnectionParameters(host=host, port=port, virtual_host=vhost)
        connection = pika.BlockingConnection(parameters)
        self.channel = connection.channel()

    def publish(self, exchange: str, queue: str, message: str) -> None:
        self.channel.basic_publish(exchange=exchange, routing_key=queue, body=message)


class WebhookTestHandler(tornado.web.RequestHandler):
    logger = logging.getLogger()

    def post(self):
        data = json.loads(self.request.body.decode("utf-8"))
        self.logger.info(data.get("status", "No status received"))
        self.logger.info(data.get("traceback", "No traceback received"))
        self.write("Received webhook")


class WebhookReceiverServer:
    def make_http_receiver_app(self):
        return tornado.web.Application(
            [
                ("/webhook", WebhookTestHandler),
            ]
        )



from types import TracebackType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import tornado.wsgi
from tornado import escape, httputil


class HurricaneWSGIException(Exception):
    pass


class HurricaneWSGIContainer(tornado.wsgi.WSGIContainer):
    """
    Wrapper for the tornado WSGI Container, which creates a WSGI-compatible function runnable on Tornado's
    HTTP server. Additionally to tornado WSGI Container should be initialized with the specific handler.

    """

    def __init__(self, handler, wsgi_application) -> None:
        self.handler = handler
        super(HurricaneWSGIContainer, self).__init__(wsgi_application)

    def _log(self, status_code: int, request: httputil.HTTPServerRequest) -> None:
        self.handler._status_code = status_code

    def __call__(self, request: httputil.HTTPServerRequest) -> None:
        data = {}  # type: Dict[str, Any]
        response = []  # type: List[bytes]

        def start_response(
            status: str,
            headers: List[Tuple[str, str]],
            exc_info: Optional[
                Tuple[
                    "Optional[Type[BaseException]]",
                    Optional[BaseException],
                    Optional[TracebackType],
                ]
            ] = None,
        ) -> Callable[[bytes], Any]:
            data["status"] = status
            data["headers"] = headers
            return response.append

        app_response = self.wsgi_application(self.environ(request), start_response)
        try:
            response.extend(app_response)
            body = b"".join(response)
        finally:
            if hasattr(app_response, "close"):
                app_response.close()  # type: ignore
        if not data:
            raise HurricaneWSGIException("WSGI app did not call start_response")

        status_code_str, reason = data["status"].split(" ", 1)
        status_code = int(status_code_str)
        headers = data["headers"]  # type: List[Tuple[str, str]]
        header_set = set(k.lower() for (k, v) in headers)
        # handle WSGI's protocol assumption the web server to strip content from HEAD requests
        # and leave content length header as is.
        # - from Django documentation:
        # Web servers should automatically strip the content of responses to HEAD requests while leaving the headers
        # unchanged, so you may handle HEAD requests exactly like GET requests in your views. Since some software,
        # such as link checkers, rely on HEAD requests, you might prefer using require_safe instead of require_GET.
        if request.method != "HEAD":
            body = escape.utf8(body)
        else:
            body = ""
        if status_code != 304:
            if "content-length" not in header_set:
                headers.append(("Content-Length", str(len(body))))
            if "content-type" not in header_set:
                headers.append(("Content-Type", "text/html; charset=UTF-8"))
        if "server" not in header_set:
            headers.append(("Server", "TornadoServer/%s" % tornado.version))

        start_line = httputil.ResponseStartLine("HTTP/1.1", status_code, reason)
        header_obj = httputil.HTTPHeaders()
        for key, value in headers:
            header_obj.add(key, value)
        if request.connection is None:
            raise ValueError("No connection")
        request.connection.write_headers(start_line, header_obj, chunk=body)
        request.connection.finish()
        self._log(status_code, request)



import asyncio

import tornado.web

from .server import metrics_log

logger = metrics_log


class K8sServerMetricsHandler(tornado.web.RequestHandler):
    """ This handler reports current usage statistics to Kubernetes """

    def get(self):
        # write custom metrics to MetricsAPI in the future
        request_queue_length = len(asyncio.all_tasks())
        logger.info(f"Request count: {request_queue_length}")
        self.write(request_queue_length)
